import socket
import struct
from .base import EMPTY_ADDR

class SOCKS4MixIn:
    version = 4
    reply_flag = 0
    code_granted = 0x5a
    code_rejected = 0x5b
    code_not_supported = 0x5b

    def pack_address(self, address = None):
        # pack_address should be unblocked so `address` MUST be (IPv4, port)
        addr = address or EMPTY_ADDR
        return struct.pack('!H4s', addr[1], socket.inet_aton(addr[0]))



import asyncio
import socket
import struct
from ..utils import parse_udp_data, SOCKS5MixIn

class UDPClient(SOCKS5MixIn):
    def __init__(self, addr):
        self.addr = addr
        self.results = asyncio.Queue()

    async def initialize(self):
        loop = asyncio.get_event_loop()
        await loop.create_datagram_endpoint(lambda : self, remote_addr=self.addr)

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, data, addr):
        data, addr = parse_udp_data(data)
        asyncio.create_task(self.results.put((data, addr)))

    def write_data(self, data, addr):
        wrapped = struct.pack('!HB', 0, 0) + self.pack_address(addr) + data
        self.transport.sendto(wrapped)



import asyncio
import io
import socket
import struct
from gera2ld.pyserve import Host
from ..utils import SOCKS5MixIn, parse_udp_data, EMPTY_ADDR
from .base import SOCKSError

class UDPRelayPeer:
    async def initialize(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind(EMPTY_ADDR)
        loop = asyncio.get_event_loop()
        await loop.create_datagram_endpoint(lambda : self, sock=sock)
        self.local_addr = self.transport.get_extra_info('sockname')[:2]
        self.len = 0

    def connection_made(self, transport):
        self.transport = transport

    def connection_lost(self, exc):
        pass

    def datagram_received(self, data, addr):
        self.len += len(data)
        self.forward_data(data, addr)

    def forward_data(self, data, addr):
        raise NotImplementedError

    def close(self):
        self.transport.close()

class UDPRelayPeerLocal(UDPRelayPeer, SOCKS5MixIn):
    client_addr = None

    def set_remote(self, remote, client_ip):
        self.remote = remote
        self.client_ip = client_ip

    def forward_data(self, data, addr):
        if self.client_ip != addr[0]:
            return
        self.client_addr = addr
        data, addr = parse_udp_data(data)
        self.remote.write_data(data, addr)
        self.len += len(data)

    def write_data(self, data):
        if self.client_addr is not None:
            self.transport.sendto(data, self.client_addr)

class UDPRelayPeerRemote(UDPRelayPeer, SOCKS5MixIn):
    def set_local(self, local):
        self.local = local

    def forward_data(self, data, addr):
        wrapped = struct.pack('!HB', 0, 0) + self.pack_address(addr) + data
        self.local.write_data(wrapped)
        self.len += len(data)

    def write_data(self, data, addr):
        self.transport.sendto(data, addr)

class UDPRelayServer:
    def __init__(self):
        self.clients = {}

    async def add_client(self, client_ip):
        local_peer = UDPRelayPeerLocal()
        remote_peer = UDPRelayPeerRemote()
        local_peer.set_remote(remote_peer, client_ip)
        remote_peer.set_local(local_peer)
        await remote_peer.initialize()
        await local_peer.initialize()
        return local_peer, remote_peer



import io
import socket
import struct
from .error import SOCKSError

__all__ = ['parse_udp_data']

def parse_udp_data(data):
    stream = io.BytesIO(data)
    _rsv, frag = struct.unpack('!HB', stream.read(3))
    assert frag == 0, 'Fragment is not supported'
    addr_type, = stream.read(1)
    if addr_type == 1:
        # IPv4
        host = socket.inet_ntoa(stream.read(4))
    elif addr_type == 3:
        # Hostname
        length, = stream.read(1)
        host = stream.read(length).decode()
    elif addr_type == 4:
        # IPv6
        host = socket.inet_ntop(socket.AF_INET6, stream.read(16))
    else:
        raise SOCKSError(f'Invalid addr_type: {addr_type}')
    port, = struct.unpack('!H', stream.read(2))
    addr = host, port
    data = stream.read()
    return data, addr



import asyncio
import socket
import struct
from ..utils import SOCKS5MixIn, forward_data
from .base import BaseHandler, SOCKSError
from .logger import logger

class SOCKS5Handler(SOCKS5MixIn, BaseHandler):
    '''
    SOCKS5 Handler
    '''
    commands = {
        1: 'connect',
        2: 'bind',
        3: 'udp',
    }

    async def reply(self, code, addr=None):
        self.writer.write(struct.pack('BBB', self.reply_flag, code, 0))
        self.writer.write(self.pack_address(addr))
        await self.writer.drain()

    async def shake_hand(self):
        length, = struct.unpack('B', (await self.reader.readexactly(1)))
        for method in struct.unpack('B' * length, (await self.reader.readexactly(length))):
            if method in self.config.socks5methods:
                self.method = method
                break
        else:
            method = self.method = 255
        self.writer.write(struct.pack('BB', self.version, method))
        await self.writer.drain()
        if method == 255:
            raise SOCKSError('No supported method')
        if method == 2:
            # Authentication needed
            res, = await self.reader.readexactly(1) # must be 0x01
            length, = struct.unpack('B', (await self.reader.readexactly(1)))
            user = await self.reader.readexactly(length)
            length, = struct.unpack('B', (await self.reader.readexactly(1)))
            pwd = await self.reader.readexactly(length)
            code = 0 if self.config.authenticate(user, pwd) else 1
            self.writer.write(struct.pack('BB', 1, code))
            await self.writer.drain()
            if code > 0:
                raise SOCKSError('Invalid user or password')
        version, command, _, addr_type = struct.unpack('BBBB', (await self.reader.readexactly(4)))
        assert version == 5, 'Invalid SOCKS version'
        if addr_type == 1:
            # IPv4
            host = socket.inet_ntoa((await self.reader.readexactly(4)))
        elif addr_type == 3:
            # Hostname
            length, = struct.unpack('B', (await self.reader.readexactly(1)))
            host = (await self.reader.readexactly(length)).decode()
        elif addr_type == 4:
            # IPv6
            host = socket.inet_ntop(socket.AF_INET6, (await self.reader.readexactly(16)))
        else:
            raise SOCKSError(f'Invalid addr_type: {addr_type}')
        port, = struct.unpack('!H', (await self.reader.readexactly(2)))
        self.addr = host, port
        logger.debug('shake_hand v5: %s %s:%d', command, host, port)
        return command

    async def socks_udp(self):
        local_peer, remote_peer = await self.udp_server.add_client(self.client_addr[0])
        self.addr = local_peer.local_addr
        await self.reply(self.code_granted, local_peer.local_addr)
        await forward_data(self.reader)
        local_peer.close()
        remote_peer.close()
        return local_peer.len, remote_peer.len, None



from .socks4 import SOCKS4Client
from .socks5 import SOCKS5Client
from ..utils import SOCKSProxy

def create_client(proxy, remote_dns=False):
    if isinstance(proxy, str):
        proxy = SOCKSProxy(proxy)
    Client = SOCKS4Client if proxy.version == 4 else SOCKS5Client
    kw = {
        'remote_dns': remote_dns,
    }
    if proxy.version == 5 and proxy.data.username is not None:
        kw['auth'] = proxy.data.username, proxy.data.password
    return Client((proxy.data.hostname, proxy.data.port), **kw)



import struct, socket
from ..utils import SOCKS5MixIn, get_host, EMPTY_ADDR
from .base import BaseClient
from .udp import UDPClient

class SOCKS5Client(SOCKS5MixIn, BaseClient):
    '''
    SOCKS5 client
    '''
    def __init__(self, addr, auth=None, remote_dns=False):
        super().__init__(addr, remote_dns)
        self.auth = auth
        self.methods = [0]
        if auth is not None:
            user, pwd = auth
            assert isinstance(user, str) and isinstance(pwd, str)
            self.methods.append(2)

    async def shake_hand(self, command, addr):
        self.writer.write(bytes([self.version, len(self.methods), *self.methods]))
        await self.writer.drain()
        version, method = struct.unpack('BB', await self.reader.readexactly(2))
        assert version == 5, 'Version unmatched'
        assert method < 255, 'Method unsupported'
        if method == 2:
            user, pwd = self.auth
            buser = user.encode()
            bpwd = pwd.encode()
            self.writer.write(bytes([1, len(buser), *buser, len(bpwd), *bpwd]))
            await self.writer.drain()
            _, ret = struct.unpack('BB', await self.reader.readexactly(2))
            assert _ == 1
            assert ret == 0, 'Authentication failed'
        if not self.remote_dns:
            addr = (await get_host(addr[0])), addr[1]
        data = struct.pack('BBB', self.version, command, 0) + self.pack_address(addr)
        self.writer.write(data)
        await self.writer.drain()

    async def load_address(self):
        _, addr_type = struct.unpack('BB', await self.reader.readexactly(2))
        if addr_type == 1:
            # IPv4
            host = socket.inet_ntoa(await self.reader.readexactly(4))
        elif addr_type == 3:
            # Hostname
            l, = struct.unpack('B', await self.reader.readexactly(1))
            host = await self.reader.readexactly(l)
        elif addr_type == 4:
            # IPv6
            host = socket.inet_ntop(socket.AF_INET6, await self.reader.readexactly(16))
        port, = struct.unpack('!H', await self.reader.readexactly(2))
        return host, port

    async def handle_udp(self):
        await self._connect()
        await self.shake_hand(3, EMPTY_ADDR)
        await self.load_reply()
        client = UDPClient(self.proxy_addr)
        await client.initialize()
        return client



import argparse
import logging
import os
from gera2ld.pyserve import run_forever
from . import SOCKSServer
from .logger import logger
from .config import Config, check_hostnames

logger.setLevel(os.environ.get('LOGLEVEL') or logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
fmt = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
ch.setFormatter(fmt)
logger.addHandler(ch)

parser = argparse.ArgumentParser(description = 'HTTP server by Gerald.')
parser.add_argument('-b', '--bind', default='127.0.0.1:1080', help='the bind address of SOCKS server')
parser.add_argument('-a', '--auth', nargs=2, action='append', help='username and password pairs')
parser.add_argument('-x', '--proxy', help='set downstream proxy')
parser.add_argument('--remote-dns', action='store_true', help='resolve DNS at remote server')
parser.add_argument('--versions', nargs='+', help='allowed versions, e.g 4 5')

args = parser.parse_args()
config = Config(args.bind, args.remote_dns)
if args.auth is not None:
    for user, pwd in args.auth:
        config.set_user(user, pwd)
if args.versions is not None:
    config.versions = set(args.versions)
if args.proxy:
    proxy = args.proxy
    if proxy == 'none': proxy = None
    config.set_proxies([
        (check_hostnames(['localhost', '127.0.0.1', '::1']), None),
        (None, proxy),
    ])
    logger.info('Use proxies: %s', config.proxies)

run_forever(SOCKSServer(config).start_server())



import asyncio
from gera2ld.pyserve import start_server_asyncio, Host
from .socks4 import SOCKS4Handler
from .socks5 import SOCKS5Handler
from .udp import UDPRelayServer
from .logger import logger

class SOCKSServer:
    handlers = {
        4: SOCKS4Handler,
        5: SOCKS5Handler,
    }

    def __init__(self, config):
        self.config = config
        self.tcp_server = None
        self.udp_server = None

    async def handle(self, reader, writer):
        try:
            version, = await reader.readexactly(1)
        except asyncio.IncompleteReadError:
            writer.close()
        else:
            await self.handle_version(reader, writer, version)

    async def handle_version(self, reader, writer, version):
        Handler = self.handlers.get(version) if version in self.config.versions else None
        if Handler is not None:
            handler = Handler(reader, writer, self.config, self.udp_server)
            name, len_local, len_remote, error = await handler.handle()
            logger.info('%s->%s %s@%d <%d >%d %s',
                Host(handler.client_addr).host,
                Host(handler.addr).host,
                name, handler.version,
                len_local, len_remote, error or '-')


    async def start_server(self):
        self.tcp_server = await start_server_asyncio(self.handle, self.config.bind, 'tcp:')
        self.udp_server = UDPRelayServer()



import asyncio

EMPTY_ADDR = '0.0.0.0', 0
BUF_SIZE = 4096

async def forward_data(reader, writer=None, count=None, timeout=None):
    while True:
        coro = reader.read(BUF_SIZE)
        if timeout is not None:
            coro = asyncio.wait_for(coro, timeout)
        data = await coro
        if not data: break
        if count is not None:
            count(len(data))
        if writer is not None:
            writer.write(data)
            await writer.drain()

class Counter:
    def __init__(self):
        self.value = 0

    def count(self, size):
        self.value += size

async def forward_pipes(reader, writer, remote_reader, remote_writer, local_timeout=None, remote_timeout=None):
    local_counter = Counter()
    remote_counter = Counter()
    task_local = asyncio.create_task(forward_data(reader, remote_writer, local_counter.count, timeout=local_timeout))
    task_remote = asyncio.create_task(forward_data(remote_reader, writer, remote_counter.count, timeout=remote_timeout))
    done, pending = await asyncio.wait(
        [task_local, task_remote],
        return_when=asyncio.FIRST_COMPLETED,
    )
    for task in pending:
        task.cancel()
    writer.close()
    remote_writer.close()
    exc = task_remote.exception() if task_remote in done else None
    return local_counter.value, remote_counter.value, exc



import asyncio
import io
import socket
import struct

class BaseClient:
    '''
    Base class of SOCKS client.
    '''
    version = NotImplemented
    reply_flag = NotImplemented
    code_granted = NotImplemented
    reader = None
    writer = None

    def __init__(self, addr, remote_dns=False):
        self.addr = addr
        self.remote_dns = remote_dns

    async def load_address(self):
        raise NotImplementedError

    async def shake_hand(self, command, addr):
        raise NotImplementedError

    async def _connect(self):
        self.reader, self.writer = await asyncio.open_connection(*self.addr)

    async def load_reply(self):
        reply_flag, code = struct.unpack('BB', await self.reader.readexactly(2))
        assert reply_flag == self.reply_flag, 'Invalid reply flag: expected %s, got %s' % (self.reply_flag, reply_flag)
        assert code == self.code_granted, 'Connection failed: expected %s, got %s' % (self.code_granted, code)
        self.proxy_addr = await self.load_address()

    async def handle_connect(self, addr):
        try:
            await self._connect()
            await self.shake_hand(1, addr)
            await self.load_reply()
        except:
            self.close()
            raise

    def close(self):
        if self.writer is not None:
            self.writer.close()
            self.reader = None
            self.writer = None



from .logger import logger
from ..utils import SOCKSProxy

def check_hostnames(rules):
    def check(host, port, hostname):
        for rule in rules:
            if rule.startswith('*.'):
                if hostname == rule[2:] or hostname.endswith(rule[1:]):
                    return True
            else:
                if hostname == rule:
                    return True
        return False
    return check

class Config:
    def __init__(self, bind='127.0.0.1:1080', remote_dns=True):
        self.users = {}
        self.versions = {5}
        self.socks5methods = 0,
        self.bind = bind
        self.set_proxies()
        self.remote_dns = remote_dns

    def set_user(self, user, pwd=None):
        if isinstance(user, str): user = user.encode()
        if isinstance(pwd, str): pwd = pwd.encode()
        if pwd is None:
            self.users.pop(user, None)
        else:
            self.users[user] = pwd
        self.socks5methods = (2,) if self.users else (0,)

    def authenticate(self, user, pwd):
        return self.users.get(user) == pwd

    def set_proxies(self, proxies=[]):
        normalized = []
        for test, proxy in proxies:
            if isinstance(proxy, str):
                proxy = SOCKSProxy(proxy)
            assert proxy is None or isinstance(proxy, SOCKSProxy)
            if isinstance(test, str):
                test = check_hostnames([test])
            assert test is None or callable(test)
            normalized.append((test, proxy))
        self.proxies = normalized

    def get_proxy(self, host, port, hostname):
        for test, proxy in self.proxies:
            if test is None or test(host, port, hostname):
                return proxy



import asyncio
from ..client import create_client
from ..utils import get_host, SOCKSError, EMPTY_ADDR, forward_pipes

class BaseHandler:
    '''
    Base handler of SOCKS protocol
    '''
    version = NotImplemented
    reply_flag = NotImplemented
    code_granted = NotImplemented
    code_rejected = NotImplemented
    code_not_supported = NotImplemented
    commands = {}

    def __init__(self, reader, writer, config, udp_server):
        self.reader = reader
        self.writer = writer
        self.config = config
        self.udp_server = udp_server
        self.client_addr = writer.get_extra_info('peername')[:2]
        self.server_addr = writer.get_extra_info('sockname')[:2]
        self.addr = '-', 0

    async def shake_hand(self):
        raise NotImplementedError

    async def reply(self, code, addr=None):
        raise NotImplementedError

    async def handle_connect_direct(self):
        host, port = self.addr
        if self.config.remote_dns:
            # Resolve host here since there is no remote proxy
            host = await get_host(host)
        return await asyncio.open_connection(host, port)

    async def handle_connect(self):
        host, port = self.addr
        hostname = host
        if not self.config.remote_dns:
            # Resolve host only if remote_dns is False
            host = await get_host(host)
        proxy = self.config.get_proxy(host=host, port=port, hostname=hostname)
        if proxy is None:
            return await self.handle_connect_direct()
        client = create_client(proxy, self.config.remote_dns)
        await client.handle_connect((host, port))
        return client.reader, client.writer

    async def socks_connect(self):
        try:
            remote_reader, remote_writer = await self.handle_connect()
        except SOCKSError:
            raise
        except Exception as e:
            if isinstance(e, OSError):
                # No route to host
                pass
            else:
                import traceback
                traceback.print_exc()
            await self.reply(self.code_rejected, EMPTY_ADDR)
            len_local = len_remote = -1
            exc = e
        else:
            await self.reply(self.code_granted, remote_writer.get_extra_info('sockname'))
            len_local, len_remote, exc = await forward_pipes(self.reader, self.writer, remote_reader, remote_writer)
        return len_local, len_remote, exc

    async def get_bind_connection(self, timeout=3):
        async def handle_bind(reader, writer):
            future.set_result((reader, writer))
        future = asyncio.Future()
        bind_server = await asyncio.start_server(handle_bind, host='0.0.0.0', port=0)
        bind_addr = bind_server.sockets[0].getsockname()[:2]
        await self.reply(self.code_granted, bind_addr)
        try:
            return await asyncio.wait_for(future, timeout)
        finally:
            bind_server.close()

    async def socks_bind(self):
        len_local = len_remote = -1
        exc = None
        try:
            remote_reader, remote_writer = await self.get_bind_connection()
        except Exception as e:
            if isinstance(e, OSError):
                pass
            else:
                import traceback
                traceback.print_exc()
            await self.reply(self.code_rejected, EMPTY_ADDR)
            exc = e
        else:
            dest_addr = remote_writer.get_extra_info('peername')
            if dest_addr[0] == self.addr[0]:
                await self.reply(self.code_granted, dest_addr)
                len_local, len_remote, exc = await forward_pipes(self.reader, self.writer, remote_reader, remote_writer)
            else:
                await self.reply(self.code_rejected, dest_addr)
        return len_local, len_remote, exc

    async def handle(self):
        command = None
        try:
            command = await self.shake_hand()
        except SOCKSError as e:
            error = e.message
        else:
            error = None
        name = self.commands.get(command)
        handle = None
        if name and error is None:
            handle = getattr(self, 'socks_' + name, None)
        len_local = len_remote = -1
        if handle is not None:
            try:
                len_local, len_remote, error = await handle()
            except asyncio.IncompleteReadError:
                pass
            except:
                raise
        else:
            await self.reply(self.code_not_supported)
        return name, len_local, len_remote, error



import asyncio
import socket
import ssl
from urllib import request
from http import client
from . import create_client
from ..utils import SOCKSProxy

class SOCKSMixIn:
    def __init__(self, *k, socks_proxy=None, **kw):
        super().__init__(*k, **kw)
        self._socks_proxy = socks_proxy

    def connect(self):
        """Connect to the host and port specified in __init__."""
        if self._socks_proxy:
            self._tunnel_socks()
        else:
            self.sock = self._create_connection(
                (self.host,self.port), self.timeout, self.source_address)
            self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        if self._tunnel_host:
            self._tunnel()

        if isinstance(self, client.HTTPSConnection):
            if self._tunnel_host:
                server_hostname = self._tunnel_host
            else:
                server_hostname = self.host

            self.sock = self._context.wrap_socket(self.sock,
                                                  server_hostname=server_hostname)
            if not self._context.check_hostname and self._check_hostname:
                try:
                    ssl.match_hostname(self.sock.getpeercert(), server_hostname)
                except Exception:
                    self.sock.shutdown(socket.SHUT_RDWR)
                    self.sock.close()
                    raise

    def _tunnel_socks(self):
        client = create_client(self._socks_proxy)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(client.handle_connect((self.host, self.port)))
        self.sock = client.writer.get_extra_info('socket')
        self.sock.setblocking(1)

class HTTPConnection(SOCKSMixIn, client.HTTPConnection):
    pass

if hasattr(client, 'HTTPSConnection'):
    class HTTPSConnection(SOCKSMixIn, client.HTTPSConnection):
        pass

class SOCKSProxyHandler(request.HTTPHandler, request.HTTPSHandler):
    def __init__(self, proxy):
        super().__init__()
        if isinstance(proxy, str):
            proxy = SOCKSProxy(proxy)
        self._socks_proxy = proxy

    def http_open(self, req):
        return self.do_open(HTTPConnection, req, socks_proxy=self._socks_proxy)

    if hasattr(client, 'HTTPSConnection'):
        def https_open(self, req):
            return self.do_open(HTTPSConnection, req,
                context=self._context, check_hostname=self._check_hostname,
                socks_proxy=self._socks_proxy)

if __name__ == '__main__':
    opener = request.build_opener(SOCKSProxyHandler('socks5://127.0.0.1:1081'))
    r = opener.open('https://www.example.com')
    print(r.read().decode())



import socket
import struct
from .base import EMPTY_ADDR

class SOCKS5MixIn:
    version = 5
    reply_flag = 5
    code_granted = 0
    code_rejected = 5
    code_not_supported = 7

    def pack_address(self, address = None):
        if address is None:
            address = EMPTY_ADDR
        host, port = address[:2]
        # IPv6
        if ':' in host:
            return struct.pack('!B16sH', 4, socket.inet_pton(socket.AF_INET6, host), port)
        # IPv4
        try:
            # IP address
            return struct.pack('!B4sH', 1, socket.inet_aton(host), port)
        except:
            # Hostname
            host = host.encode()
            return struct.pack('!BB', 3, len(host)) + host + struct.pack('!H', port)



from urllib.parse import urlparse, urlunparse

class SOCKSProxy:
    def __init__(self, bind):
        self.data = urlparse(bind)
        assert self.data.scheme in ('socks4', 'socks5'), 'Invalid SOCKS version'
        self.version = int(self.data.scheme[-1])

    def __repr__(self):
        return '<' + str(self) + '>'

    def __str__(self):
        return urlunparse(self.data)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return hash(self) == hash(other)



import asyncio
import os
import socket
from async_dns import types
from async_dns.resolver import ProxyResolver

resolver = None

def set_resolver(_resolver=None):
    global resolver
    resolver = _resolver or ProxyResolver(proxies=[
        (None, os.environ.get('GERA2LD_SOCKS_NAMESERVER', '114.114.114.114').split(',')),
    ])

def is_ip(host):
    try:
        socket.inet_pton(socket.AF_INET6 if ':' in host else socket.AF_INET, host)
    except OSError:
        return False
    return True

async def get_host(host, qtypes=(types.A, types.AAAA)):
    if is_ip(host):
        return host
    if resolver is None:
        set_resolver()
    for qtype in qtypes:
        try:
            res = await resolver.query(host, qtype)
        except asyncio.IncompleteReadError:
            pass
        else:
            for item in res.an:
                if item.qtype in (types.A, types.AAAA):
                    ip = item.data
                    if ip: return ip
    raise Exception(f'DNS lookup failed: {host}')



import asyncio
import socket
import struct
from ..utils import SOCKS4MixIn, get_host
from .base import BaseClient

class SOCKS4Client(SOCKS4MixIn, BaseClient):
    '''
    SOCKS4 client
    '''
    def __init__(self, addr, userid='', remote_dns=False):
        super().__init__(addr, remote_dns)
        self.userid = userid.strip('\0')

    async def shake_hand(self, command, addr):
        self.writer.write(struct.pack('B', self.version))
        hostname = addr[0]
        # SOCKS4a
        remote_dns = self.remote_dns
        try:
            socket.inet_aton(hostname)
            remote_dns = False
        except OSError:
            if remote_dns:
                addr = '0.0.0.1', addr[1]
            else:
                addr = (await get_host(addr[0])), addr[1]
        data = struct.pack('B', command) + self.pack_address(addr) + self.userid.encode() + b'\0'
        if remote_dns:
            data += hostname.encode() + b'\0'
        self.writer.write(data)
        await self.writer.drain()

    async def load_address(self):
        port, = struct.unpack('!H', await self.reader.readexactly(2))
        ipn = await self.reader.readexactly(4)
        ip = socket.inet_ntoa(ipn)
        return ip, port



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gera2ld',
 'gera2ld.socks',
 'gera2ld.socks.client',
 'gera2ld.socks.server',
 'gera2ld.socks.utils']

package_data = \
{'': ['*']}

install_requires = \
['async_dns>=1.1.3,<2.0.0', 'gera2ld-pyserve>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'gera2ld.socks',
    'version': '0.4.5',
    'description': 'SOCKS client and server based on asyncio',
    'long_description': "# gera2ld.socks\n\n[![PyPI](https://img.shields.io/pypi/v/gera2ld.socks.svg)](https://pypi.org/project/gera2ld.socks/)\n\nThis is a SOCKS server and client package built with `asyncio` (requires Python 3.5+).\n\n## Installation\n\n``` sh\n$ pip3 install gera2ld.socks\n```\n\n## Usage\n\n* SOCKS server\n\n  shell command:\n  ``` sh\n  # Start a server\n  $ python3 -m gera2ld.socks.server -b 127.0.0.1:1080\n  ```\n\n  or python script:\n  ``` python\n  from gera2ld.pyserve import run_forever\n  from gera2ld.socks.server import Config, SOCKSServer\n\n  config = Config('127.0.0.1:1080')\n  run_forever(SOCKSServer(config).start_server())\n  ```\n\n* SOCKS client\n\n  ``` python\n  import asyncio\n  from gera2ld.socks.client import create_client\n\n  client = create_client('socks5://127.0.0.1:1080')\n  loop = asyncio.get_event_loop()\n  loop.run_until_complete(client.handle_connect(('www.google.com', 80)))\n  client.writer.write(b'...')\n  print(loop.run_until_complete(client.reader.read()))\n  ```\n\n* SOCKS handler for `urllib`\n\n  ``` python\n  from urllib import request\n  from gera2ld.socks.client.handler import SOCKSProxyHandler\n\n  handler = SOCKSProxyHandler('socks5://127.0.0.1:1080')\n\n  opener = request.build_opener(handler)\n  r = opener.open('https://www.example.com')\n  print(r.read().decode())\n  ```\n\n* SOCKS client for UDP\n\n  ``` python\n  import asyncio\n  from gera2ld.socks.client import create_client\n\n  async def main():\n      client = create_client('socks5://127.0.0.1:1080')\n      udp = await client.handle_udp()\n      udp.write_data(b'\\xc9\\xa7\\x01\\x00\\x00\\x01\\x00\\x00\\x00\\x00\\x00\\x00\\x03www\\x06google\\x03com\\x00\\x00\\xff\\x00\\x01', ('114.114.114.114', 53))\n      print(await udp.results.get())\n\n  loop = asyncio.get_event_loop()\n  loop.run_until_complete(main())\n  ```\n",
    'author': 'Gerald',
    'author_email': 'gera2ld@163.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gera2ld/pysocks',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import io
import socket
import struct
from ..utils import SOCKS4MixIn
from .base import BaseHandler
from .logger import logger

class SOCKS4Handler(SOCKS4MixIn, BaseHandler):
    '''
    SOCKS4 Handler
    '''
    commands = {
        1: 'connect',
        2: 'bind',
    }

    async def reply(self, code, addr=None):
        self.writer.write(struct.pack('BB', self.reply_flag, code))
        # address must be IPv4
        self.writer.write(self.pack_address(addr))
        await self.writer.drain()

    async def shake_hand(self):
        command, port = struct.unpack('!BH', (await self.reader.readexactly(3)))
        ip = await self.reader.readexactly(4)
        ipn, = struct.unpack('!I', ip)
        # SOCKS4a supports domain names by given IP 0.0.0.x (x > 0)
        socks4a = ipn < 256
        userid = (await self.reader.readuntil(b'\0'))[:-1]
        if socks4a:
            host = (await self.reader.readuntil(b'\0'))[:-1]
        else:
            host = socket.inet_ntoa(ip)
        self.addr = host, port
        logger.debug('shake_hand v4: %s %s:%d', command, host, port)
        return command
